 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Debugger
  * 
  * (c) 1995 Bernd Schmidt
  * 
  */

#include "sysconfi.h"
#include "sysdeps.h"

#include <ctype.h>
#include <signal.h>
#include <conio.h>

#include "config.h"
#include "options.h"
#include "debug.h"
#include "memory2.h"
#include "Memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "r_newcpu.h"
#include "ram.h"
#include "prefs.h"

extern int quit_program;
extern int use_debugger;

static CPTR skipaddr;
static int do_skip;

int debugger_active = 0;
int _debugging = 0;
int Insert_Disk=0;
int go=0;

// prototypes
void InsertDisk (void);

static RETSIGTYPE sigbrkhandler(int foo)
{
   extern void EnterDebug(void);

   EnterDebug();

   do_skip = 0;
   debugger_active = 1;
   _debugging = 1;
 
   broken_in = 1;
   regs.spcflags |= SPCFLAG_BRK;
   signal(SIGINT, sigbrkhandler);
}

static void ignore_ws(char **c)
{
   while (**c && isspace(**c))
      (*c)++;
}

static ULONG readhex(char **c)
{
   ULONG val = 0;
   char nc;

   ignore_ws(c);

   while (isxdigit(nc = **c)) {
      (*c)++;
      val *= 16;
      nc = toupper(nc);
      if (isdigit(nc))
         val += nc - '0';
      else
          val += nc - 'A' + 10;
   }

   return val;
}

static char next_char(char **c)
{
   ignore_ws(c);
   return *(*c)++;
}

static int more_params(char **c)
{
   ignore_ws(c);
   return (**c) != 0;
}

static void dumpmem(CPTR addr, CPTR *nxmem, int lines)
{
   broken_in = 0;
   for (;lines-- && !broken_in;) {
      int i, j;

      cprintf("%08lx    ", addr);

      for (i=0; i<8; i++) {
         cprintf("%04x ", get_word(addr));
         j=wherex();
         gotoxy(56+(i*2), wherey());
         if (get_byte(addr)==0 || get_byte(addr)==7 || get_byte(addr)==8 ||
             get_byte(addr)==9 || get_byte(addr)==10|| get_byte(addr)==13||
             get_byte(addr)==32|| get_byte(addr)==255)
            cprintf(".");
         else
            cprintf("%c", (char) get_byte(addr));

         if (get_byte(addr+1)==0 || get_byte(addr+1)==7 || get_byte(addr+1)==8 ||
             get_byte(addr+1)==9 || get_byte(addr+1)==10|| get_byte(addr+1)==13||
             get_byte(addr+1)==32|| get_byte(addr+1)==255)
            cprintf(".");
         else
            cprintf("%c", (char) get_byte(addr+1));

         gotoxy(j, wherey());
         addr+=2;
      }

      printf("\n");
   }
   *nxmem = addr;
}


void InsertDisk(void)
{
   UWORD  drivenum=Insert_Disk;
   char   diskfile[255];

   extern int Sony_Disk_Inserted (UWORD Drive_No);
   extern void Sony_Insert (UWORD Drive_No, char *filename);
                      
   printf("enter filedisk name to mount in drive %d: ", drivenum);
   scanf("%255s", &diskfile);
   fflush(stdin);

   if (!Sony_Disk_Inserted(drivenum)) {
      Sony_Insert(drivenum, (char *)diskfile);                            
      if (Sony_Disk_Inserted(drivenum)) {
         strcpy(thePrefs.DrivePath[drivenum-1], diskfile);
      }
   } else {
      printf("There already is a disk in drive %d.\n", drivenum);
   }
}


void debug(void)
{
   char input[80];
   CPTR nextpc,nxdis,nxmem;

   signal(SIGINT, sigbrkhandler);

   if (!use_debugger) {
      go=1;
   } else {
      raise(SIGINT);

      printf("debugging...\n");
      do_skip = 0;

      m68k_dumpstate(&nextpc);
      nxdis = nextpc; nxmem = 0;
   }

   for (;;) {
      char cmd,*inptr;

      if (!go) {
         printf(">");
         fgets(input, 80, stdin);
         inptr = input;
         cmd = next_char(&inptr);
      } else {
         go=0;
         cmd = 'g';
      }

      switch(cmd) {
         case 'w': {
            ULONG maddr;
            int value;

            if (more_params(&inptr))
               maddr = readhex(&inptr); 
            else 
               break;

            if (more_params(&inptr))
               value = readhex(&inptr); 
            else
               break;

            put_byte(maddr, value);

            break;
         }

         case 'D': {
            FILE *RAMFile;
            char *MacRAM = (char *) get_real_address (kRAM_Size-0x5900);
            unsigned long i;

            RAMFile = fopen("vMac.SCR", "wb+");

            fwrite(MacRAM, (512*342/8), 1, RAMFile);
            fclose(RAMFile);
                    
            break;
         }

         case 'q': {
            return;
         }

         case 'r': {
            m68k_dumpstate(&nextpc);
            break;
         }

         case 'd': {
            ULONG daddr;
            int count;
		
            if (more_params(&inptr))
               daddr = readhex(&inptr);
            else 
               daddr = nxdis;

            if (more_params(&inptr))
               count = readhex(&inptr); 
            else
               count = 10;

            m68k_disasm(daddr, &nxdis, count);
            break;
         }

         case 'R': {
            extern void Reset_Emulator(void);

            Reset_Emulator();

            break;
         }

         case 'I': {
            extern int InterruptNo;
	   
            regs.spcflags |= SPCFLAG_DOINT;
            InterruptNo = 4;
            break;
         }

         case 'g': {
            extern int InstallGraphics(void);

            if (use_debugger) {
               if (more_params (&inptr))
                  m68k_setpc (readhex (&inptr));
            }

            debugger_active = 0;
            _debugging = 0;

            InstallGraphics();

            regs.spcflags &= ~SPCFLAG_BRK;
            quit_program=0;

            m68k_go(1);
            signal(SIGINT, sigbrkhandler);
            raise(SIGINT);

            if (Insert_Disk>0) {
               InsertDisk();
               Insert_Disk=0;
               go=1;
            } else {
               printf("debugging...\n");
               m68k_dumpstate(&nextpc);
               nxdis = nextpc; nxmem = 0;
            }
            break;
         }

         case 't': {
            reentrant_m68k_step();
            m68k_dumpstate(&nextpc);
            break;
         }

         case 'f': {
            CPTR stop;

            stop = readhex(&inptr);
            while (stop!=nextpc) {
               reentrant_m68k_step();
            }

            break;
         }

         case 'm': {
            ULONG maddr;
            int lines;

            if (more_params(&inptr))
               maddr = readhex(&inptr); 
            else 
               maddr = nxmem;

            if (more_params(&inptr))
               lines = readhex(&inptr); 
            else 
               lines = 16;

            dumpmem(maddr, &nxmem, lines);

            break;
         }

         case 'c': {
            FILE *RAMFile;
            char *MacRAM = (char *) get_real_address (kRAM_Size-kRAM_Size);
            unsigned long i;

            RAMFile = fopen("vMac.RAM", "wb+");
            fwrite(MacRAM, kRAM_Size, 1, RAMFile);
            fclose(RAMFile);
                    
            break;
         }  

         case 'i': {
            extern unsigned long int Instrucs;

            printf("vMac: Instruction_counter: %ld\n", thePrefs.Instrucs);
            printf("vMac: estimated m68k CPU Speed: %ld.%ld mHz.\n", ((thePrefs.Instrucs*165) / 1000000), ((thePrefs.Instrucs*165) % 1000000));

            break;
         }

         case 'h':
         case '?': {
            printf ("          HELP for vMac Debugger\n");
            printf ("         -----------------------\n\n");
            printf ("  h,?:                  Show this help page                                   \n");
            printf ("  g: <address>          Start execution at the current address or <address>.  \n");
            printf ("  r:                    Dump state of the CPU                                 \n");
            printf ("  m <address> <lines>:  Memory dump starting at <address>                     \n");
            printf ("  d <address> <lines>:  Disassembly starting at <address>                     \n");
            printf ("  w <address> <value>:  Write a value (byte size) at <address>                \n");
            printf ("  t:                    Step one instruction                                  \n");
            printf ("  f <address>:          Step forward until PC == <address>                    \n");
            printf ("  c:                    Dump the RAM contents into ~vMac.RAM~.                \n");
            printf ("  D:                    Dump the screen into ~vMAC.SCR~. (raw data)           \n");
            printf ("  R:                    Restart the Macintosh Plus.                           \n");
            printf ("  I:                    Interrupt the Macintosh Plus.                         \n");
            printf ("  q:                    Quit the emulator.                                  \n\n");
            break;
         }
      }
   }
}
