/*
   vMac - Screen.C 
   screen interface
   (rewritten from version 0.1.8.2)

   (c) Gustavo Goedert, 1997
   (c) Yoav Shadmi, 1997-1998
*/

#include "sysconfi.h"
#include "sysdeps.h"

#include <dos.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/farptr.h>
#include <crt0.h>
#include <ctype.h>
#include <signal.h>
#include <sys/fsext.h>
#include <stdlib.h>
#include <conio.h>
#include <sys/time.h>
#include <string.h>
#include <stdio.h>

#include "config.h"

#include "Memory.h"
#include "memory2.h"

#include "custom2.h"
#include "readcpu.h"
#include "newcpu.h"

#include "Screen.h"
#include "RAM.h"
#include "VIA.h"
#include "Keyboard.h"
#include "Sony.h"
#include "prefs.h"
#include "debug.h"
#include "interrup.h"
#include "universa.h"

#include "keys.h"

#define MacWidth        512
#define MacHeight       342
#define RegScreen       0xa1598
#define BitsPerPixel    1
#define Depth           (8/BitsPerPixel)
#define MacSize         MacWidth*MacHeight/Depth

#define SC_ESC 0xFFFE
#define SC_CTR 0xFFFD
#define SC_ALT 0xFFFC
#define SC_SFT 0xFFFB
#define SC_CLK 0xFFFA

// screen
int no_msg=0;
UBYTE vPage2 = 1;
char  *MacVideo, *VGAVideo;
int OldVGAMode;
int InGraphics = 0;
int UseScreen = 0;
extern int Insert_Disk;

extern int quit_program;

// keyboard & mouse
extern int _quit_;
int KeyBoardInstalled = 0;
int MouseInstalled = 0;
extern volatile int buttonstate;
volatile int newmousecounters = 0;
volatile int lastmx = 0, lastmy = 0;

int MacKeys[128] = {
   0xFFFF, SC_ESC, 0x1231, 0x1332, 0x1433, 0x1534, 0x1735, 0x1636,
   0x1A37, 0x1C38, 0x1939, 0x1D30, 0x1B2D, 0x183D, 0x3308, 0x3009,
   0x0C71, 0x0D77, 0x0E65, 0x0F72, 0x1174, 0x1079, 0x2075, 0x2269,
   0x1F6F, 0x2370, 0xFFFF, 0xFFFF, 0x490D, SC_CTR, 0x0061, 0x0173,
   0x0264, 0x0366, 0x0567, 0x0468, 0x266A, 0x286B, 0x256C, 0x293B,
   0x2727, 0xB27E, SC_SFT, 0x2A5C, 0x067A, 0x0778, 0x0863, 0x0976,
   0x0B62, 0x2D6E, 0x2E6D, 0x2B2C, 0x2F2E, 0x2C2F, SC_SFT, 0x9C2A,
   SC_ALT, 0x3120, SC_CLK, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x1A37,
   0x1C38, 0x1939, 0x1B2D, 0x1534, 0x1735, 0x1636, 0x982B, 0x1231,
   0x1332, 0x1433, 0x1D30, 0x2F2E, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0x490D, SC_CTR, 0x2C2F, 0xFFFF, SC_ALT, 0xFFFF, 0xFFFF, 0x4D1E,
   0xFFFF, 0x461C, 0x421D, 0xFFFF, 0x481F, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
};

unsigned char EscapeKeys[128] = {
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   96,  97,  0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   98,  0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   99,
    100, 0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   101, 102,
    103, 104, 0,   105, 0,   106, 0,   107,
    108, 109, 110, 111, 0,   0,   0,   0,
    0,   0,   0,   125, 126, 127, 0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
};

_go32_dpmi_seginfo OldKeyboardHandler, NewKeyboardHandler;
_go32_dpmi_seginfo NewMouseHandler;
_go32_dpmi_registers MouseCallbackRegs;

int KeyState[256];

void Quit_Emulator(void);
static void KeyboardHandler(void);
static void MouseHandler(_go32_dpmi_registers *MouseRegs);

// timer
static RETSIGTYPE TimerHandler(int foo);
struct itimerval timer;

void RemapColors (int set);

int  InstallGraphics(void);
int  InstallMouse(void);
int  InstallTimer(void);
int  InstallKeyboard(void);
void UninstallMouse(void);
void UninstallTimer(void);
void UninstallKeyboard(void);

volatile int ClockTicks, OldTimer, NewTick = 0;
_go32_dpmi_seginfo OldTimerHandler, NewTimerHandler;

#define MAXMESSAGES 100
char *Messages[MAXMESSAGES];
int CurrentMessage = 0;

// Prototypes.
extern void Quit_vMac(void);
extern void m68k_reset(void);
static int ScreenWrite(__FSEXT_Fnumber func, int *retval, va_list rest_args);
void   DumpMessages(void);
void EnterDebug(void);
void Reset_Emulator(void);

inline void Screen_Draw (void)
{
   char    *VGAVideo, *MacVideo;
   int     i;
   UWORD   Word, PrevWord;

   if (vPage2 == 1)
      MacVideo = (char *) RAM_Xlate(kMain_Buffer);
   else
      MacVideo = (char *) RAM_Xlate(kAlternate_Buffer);

   (char *)VGAVideo = (char *)RegScreen;

   for (i=0; i<342; i++, MacVideo+=64, VGAVideo+=80)
      asm __volatile__ ("
              movw    %0, %%fs
      NextPixel:
              movl    (%%esi), %%eax
              notl    %%eax  
              movl    %%eax, %%fs:(%%edi)
              addl    $4, %%esi
              addl    $4, %%edi
              loop    NextPixel
              "
              :
              : "g" (_dos_ds), "S" (MacVideo), "D" (VGAVideo), "c" (16)
              : "%eax", "%ebx", "%edx", "%esi", "%edi", "%ecx", "cc", "memory"
      );
}

int Screen_Init(int argc, char **argv)
{
   int i;
   _go32_dpmi_registers Regs;

   /* whenever we try to exit the application, Quit_Emulator will be called */
   atexit(Quit_Emulator);

   /* redirect output */
   __FSEXT_set_function(fileno (stdout), ScreenWrite);
   __FSEXT_set_function(fileno (stderr), ScreenWrite);
   for (i=0; i<MAXMESSAGES; i++)
      Messages[i] = NULL;

   /* Save Old VGA Mode */
   Regs.x.ax=0x1C;
   Regs.h.ah=0x0F;
   Regs.x.ss=Regs.x.sp=Regs.x.flags=0;
   _go32_dpmi_simulate_int(0x10, &Regs);
   OldVGAMode = Regs.h.al;

   /* start MS-DOS event handling. */
   if (!InstallMouse())
      return 1;
/*
   if (!InstallTimer())
      return 1;
*/
   if (!InstallGraphics())
      return 1;

   _go32_dpmi_lock_code(Screen_Draw, (unsigned int)Screen_Init-(unsigned int)Screen_Draw);

   return 0;
}

void Screen_Reset(void)
{
}

int InstallGraphics (void)
{
   _go32_dpmi_registers Regs;

   Regs.h.ah = 0x00;
   Regs.h.al = 0x11;
   Regs.x.ss = Regs.x.sp = Regs.x.flags = 0;
   _go32_dpmi_simulate_int(0x10, &Regs);

   InstallKeyboard();
   InGraphics = 1;
   UseScreen = 0;
   return 1;
}

void DumpMessages(void)
{
   int i, StartMessage, MessagePosition;

   UseScreen = 1;
   if (CurrentMessage) {
      StartMessage = CurrentMessage - MAXMESSAGES;
      if (StartMessage < 0)
         StartMessage = 0;

      for (i=StartMessage; i<CurrentMessage; i++) {
         MessagePosition = i % MAXMESSAGES;
         if (Messages[MessagePosition] != NULL)
            printf(Messages[MessagePosition]);
      }
      CurrentMessage = 0;
   }
}

static int ScreenWrite(__FSEXT_Fnumber func, int *retval, va_list rest_args)
{
   char *buf, *mybuf;
   size_t buflen;
   int fd = va_arg (rest_args, int);
   int MessagePosition;

   if (func != __FSEXT_write || UseScreen)
      return 0;

   buf = va_arg(rest_args, char *);
   buflen = va_arg(rest_args, size_t);
   mybuf = malloc(buflen + 1);

   if (mybuf != NULL) {
      memcpy (mybuf, buf, buflen);
      mybuf[buflen] = '\0';
   }

   MessagePosition = CurrentMessage % MAXMESSAGES;
   if (Messages[ MessagePosition ] != NULL)
      free(Messages[ MessagePosition ]);

   Messages[ MessagePosition ] = mybuf;

   CurrentMessage++;

   *retval = buflen;
   return 1;
}

void EnterDebug(void)
{
   _go32_dpmi_registers Regs;

   if (InGraphics == 0)
      return;
   InGraphics = 0;

   Regs.x.ax = 0x1C;
   Regs.h.ah = 0x00;
   Regs.h.al = 0x03;
   Regs.x.ss = Regs.x.sp = Regs.x.flags = 0;
   _go32_dpmi_simulate_int(0x10, &Regs);

   UninstallKeyboard();
   _set_screen_lines(50);
   clrscr();
   DumpMessages();
}

// keyboard
static void KeyboardHandler(void)
{
   extern int use_debugger;

   static int IsEscape = 0, CapsLockCanChange = 1;
   int ScanCode, NewState, MacKey, Mods;

   ScanCode = inportb(0x60);

   if (ScanCode == 0xe0) {
      IsEscape = 1;
      outportb(0x20, 0x20);
      return;
   }

   // PAUSE Key
   if (IsEscape > 1) {
      IsEscape++;
      if (IsEscape == 7)
         IsEscape = 0;
      outportb(0x20, 0x20);
      return;
   }

   if (ScanCode == 0xe1) {
      IsEscape = 2;
      outportb(0x20, 0x20);
      return;
   }

   NewState = !(ScanCode & 0x80);
   ScanCode = ScanCode & 0x7f;

   if (IsEscape) {
      ScanCode = EscapeKeys[ScanCode];
      IsEscape = 0;
   }

   if (ScanCode==thePrefs.INSERT1) {
      if (!Sony_Disk_Inserted(1)) {
         Sony_Insert(0x0001, "A:");
      }
   }
   else if (ScanCode==thePrefs.INSERT2) {
      if (!Sony_Disk_Inserted(2)) {
            Insert_Disk=2;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
      }
   }
   else if (ScanCode==thePrefs.INSERT3) {
      if (!Sony_Disk_Inserted(3)) {
            Insert_Disk=3;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
      }
   }
   else if (ScanCode==thePrefs.INSERT4) {
      if (!Sony_Disk_Inserted(4)) {
            Insert_Disk=4;
            regs.spcflags |= SPCFLAG_BRK;
            use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
            quit_program = 1;
      }
   }
   else if (ScanCode==thePrefs.DEBUG) {
         regs.spcflags |= SPCFLAG_BRK;
         use_debugger = 1;       /* 1 - go to debug ; 0 - exit */
         quit_program = 1;
   }
   else if (ScanCode==thePrefs.EXIT) {
      regs.spcflags |= SPCFLAG_BRK;
      use_debugger = 0;       /* 1 - go to debug ; 0 - exit */
      quit_program = 1;
   }
   else if (ScanCode==thePrefs.INTERRUPT) {
      extern int InterruptNo;
      if (!use_debugger)
         use_debugger=1;
      regs.spcflags |= SPCFLAG_DOINT;
      InterruptNo = 4;   
   }    
   else if (ScanCode==thePrefs.RESET) {
      if (!use_debugger)
         use_debugger=1;
      Reset_Emulator();
   }  

   outportb(0x20, 0x20);

   MacKey = MacKeys[ScanCode];

   if (MacKey == 0xFFFF)
      return;

   if (MacKey == SC_CLK) {
      if (NewState) {
         if (CapsLockCanChange) {
            NewState = !KeyState[ScanCode];
            CapsLockCanChange = 0;
         } else
            return;
      } else {
         CapsLockCanChange = 1;
         return;
      }
   }

   if (NewState && KeyState[ScanCode])
       NewState = 2;
   KeyState[ScanCode] = NewState;

   if (MacKey < 0xFF00) {
      Mods=0;
      switch (thePrefs.COMMAND) {
         case 0: if (KeyState[56] || KeyState[100]) Mods |= 0x0100; break;
         case 1: if (KeyState[56]) Mods |= 0x0100; break;
         case 2: if (KeyState[100]) Mods |= 0x0100; break;
         case 3: if (KeyState[29] || KeyState[97]) Mods |= 0x0100; break;
         case 4: if (KeyState[29]) Mods |= 0x0100; break;
         case 5: if (KeyState[79]) Mods |= 0x0100; break;
      }   

      if (KeyState[58]) {                      // Caps-Lock
         Mods |= 0x0400;
         if (((MacKey & 0x00FF)>=97) && ((MacKey & 0x00FF)<=122))
            MacKey-=32;                        // Capital Letters
      }
      if (KeyState[42] || KeyState[54]) {      // Shift
         Mods |= 0x0200;
         if (((MacKey & 0x00FF)>=97) && ((MacKey & 0x00FF)<=122)) {
            MacKey-=32;
         } else {
            switch (MacKey & 0x00FF) {
               /*`*/ case 0x60: MacKey = (MacKey&0xFF00)+0x7E; break;
               /*1*/ case 0x31: MacKey = (MacKey&0xFF00)+0x21; break;
               /*2*/ case 0x32: MacKey = (MacKey&0xFF00)+0x40; break;
               /*3*/ case 0x33: MacKey = (MacKey&0xFF00)+0x23; break;
               /*4*/ case 0x34: MacKey = (MacKey&0xFF00)+0x24; break;           
               /*5*/ case 0x35: MacKey = (MacKey&0xFF00)+0x25; break;
               /*6*/ case 0x36: MacKey = (MacKey&0xFF00)+0x5E; break;
               /*7*/ case 0x37: MacKey = (MacKey&0xFF00)+0x26; break;
               /*8*/ case 0x38: MacKey = (MacKey&0xFF00)+0x2A; break;
               /*9*/ case 0x39: MacKey = (MacKey&0xFF00)+0x28; break;
               /*0*/ case 0x30: MacKey = (MacKey&0xFF00)+0x29; break;
               /*-*/ case 0x2D: MacKey = (MacKey&0xFF00)+0x5F; break;
               /*=*/ case 0x3D: MacKey = (MacKey&0xFF00)+0x2B; break;
               /*\*/ case 0x5C: MacKey = (MacKey&0xFF00)+0x7C; break;
               /*[*/ case 0x5B: MacKey = (MacKey&0xFF00)+0x7B; break;
               /*]*/ case 0x5D: MacKey = (MacKey&0xFF00)+0x7D; break;
               /*;*/ case 0x3B: MacKey = (MacKey&0xFF00)+0x3A; break;
               /*'*/ case 0x27: MacKey = (MacKey&0xFF00)+0x22; break;
               /*,*/ case 0x2C: MacKey = (MacKey&0xFF00)+0x3C; break;
               /*.*/ case 0x2E: MacKey = (MacKey&0xFF00)+0x3E; break;
               /*?*/ case 0x2F: MacKey = (MacKey&0xFF00)+0x3F; break;
            }
         }
      }

      switch (thePrefs.OPTION) {
         case 0: if (KeyState[56] || KeyState[100]) Mods |= 0x0800; break;
         case 1: if (KeyState[56]) Mods |= 0x0800; break;
         case 2: if (KeyState[100]) Mods |= 0x0800; break;
         case 3: if (KeyState[29] || KeyState[97]) Mods |= 0x0800; break;
         case 4: if (KeyState[29]) Mods |= 0x0800; break;
         case 5: if (KeyState[79]) Mods |= 0x0800; break;
      }   

      switch (NewState) {
         case 0: PC_PostEvent (4, MacKey, Mods); break;
         case 1: PC_PostEvent (3, MacKey, Mods); break;
         case 2: PC_PostEvent (5, MacKey, Mods); break;
      }
   }
}

int InstallKeyboard(void)
{
   int i;

   if (KeyBoardInstalled)
      return(1);

   _go32_dpmi_get_protected_mode_interrupt_vector(9, &OldKeyboardHandler);

   NewKeyboardHandler.pm_offset = (int)KeyboardHandler;
   if (_go32_dpmi_allocate_iret_wrapper(&NewKeyboardHandler) != 0) {
       EnterDebug();
      _set_screen_lines(25);
      _quit_=1;
      fprintf (stderr, "vMac: Can't allocate keyboard iret_wrapper.\n");
      raise(SIGUSR1);
      return(0);
   }

   if (_go32_dpmi_set_protected_mode_interrupt_vector(9, &NewKeyboardHandler) != 0) {
       EnterDebug();
      _set_screen_lines(25);
      _quit_=1;
      fprintf(stderr, "Can't set protected mode interrupt vector.\n");
      raise(SIGUSR1);
      return(0);
   }

   for (i = 0; i < 256; i++)
      KeyState[i] = 0;

   _go32_dpmi_lock_code(KeyboardHandler, (unsigned int)InstallKeyboard-(unsigned int)KeyboardHandler);
   _go32_dpmi_lock_data(KeyState, sizeof(KeyState));
   _go32_dpmi_lock_data(&OldKeyboardHandler, sizeof(OldKeyboardHandler));
   _go32_dpmi_lock_data(&NewKeyboardHandler, sizeof(NewKeyboardHandler));

   KeyBoardInstalled = 1;

   return(1);
}

void UninstallKeyboard(void)
{
   if (!KeyBoardInstalled)
      return;
   _go32_dpmi_set_protected_mode_interrupt_vector(9, &OldKeyboardHandler);
   _go32_dpmi_free_iret_wrapper(&NewKeyboardHandler);
   KeyBoardInstalled = 0;
}


// mouse
static void MouseHandler(_go32_dpmi_registers *MouseRegs)
{
   lastmx = (short)MouseRegs->x.cx;      //si
   lastmy = (short)MouseRegs->x.dx;      //di
   buttonstate = MouseRegs->x.bx & 7;
   newmousecounters = 1;
}

int InstallMouse(void)
{
    _go32_dpmi_registers IntRegs;

    if (MouseInstalled)
	return(1);
    IntRegs.x.ax = 0;
    IntRegs.x.ss = IntRegs.x.sp = IntRegs.x.flags = 0;
    _go32_dpmi_simulate_int(0x33, &IntRegs);
    if (IntRegs.x.ax==0) {
       EnterDebug();
       _set_screen_lines(25);
       _quit_=1;
       fprintf(stderr, "vMac: Mouse driver not present.\n");
       raise(SIGUSR1);
       return(0);
    }

    NewMouseHandler.pm_offset = (int)MouseHandler;
    if (_go32_dpmi_allocate_real_mode_callback_retf(&NewMouseHandler, &MouseCallbackRegs) != 0) {
       EnterDebug();
       _set_screen_lines(25);
       _quit_=1;
       fprintf(stderr, "vMac: Can't allocate mouse callback_retf.\n");
       raise(SIGUSR1);
       return(0);
    }
    IntRegs.x.ax = 0xc;
    IntRegs.x.cx = 0x7f;
    IntRegs.x.es = NewMouseHandler.rm_segment;
    IntRegs.x.dx = NewMouseHandler.rm_offset;
    IntRegs.x.ss = IntRegs.x.sp = IntRegs.x.flags = 0;
    buttonstate = 0;
    lastmx = lastmy = 0;
    _go32_dpmi_simulate_int(0x33, &IntRegs);
    _go32_dpmi_lock_code(MouseHandler, (unsigned int)InstallMouse-(unsigned int)MouseHandler);
    _go32_dpmi_lock_data(&lastmx, sizeof(lastmx));
    _go32_dpmi_lock_data(&lastmy, sizeof(lastmy));
    _go32_dpmi_lock_data(&buttonstate, sizeof(buttonstate));
    _go32_dpmi_lock_data(&newmousecounters, sizeof(newmousecounters));
    _go32_dpmi_lock_data(&NewMouseHandler, sizeof(NewMouseHandler));
    _go32_dpmi_lock_data(&MouseCallbackRegs, sizeof(MouseCallbackRegs));
    MouseInstalled = 1;
    return(1);
}

void UninstallMouse(void)
{
    _go32_dpmi_registers IntRegs;

    if (!MouseInstalled)
	return;
    IntRegs.x.ax = 0xc;
    IntRegs.x.cx = 0x0;
    IntRegs.x.ss = IntRegs.x.sp = IntRegs.x.flags = 0;
    _go32_dpmi_simulate_int(0x33, &IntRegs);
    _go32_dpmi_free_real_mode_callback(&NewMouseHandler);
    MouseInstalled = 0;
}

// timer
static RETSIGTYPE TimerHandler(int foo)
{
   extern void Int_Handler (void);

   Int_Handler();
}

int InstallTimer(void)
{
   // activate timer and signal handling
   signal(SIGPROF, TimerHandler);
   timer.it_interval.tv_sec  = 0;
   timer.it_interval.tv_usec = 16625;
   timer.it_value.tv_sec     = 0;
   timer.it_value.tv_usec    = 16625;
   setitimer(ITIMER_PROF, &timer, NULL);

   _go32_dpmi_lock_code(TimerHandler, (unsigned int)InstallTimer-(unsigned int)TimerHandler);
   _go32_dpmi_lock_data(&timer, sizeof(timer));

   return 1;
}

void UninstallTimer(void)
{
   // de-activate timer and signal handling
   signal(SIGPROF, SIG_DFL);
   timer.it_interval.tv_sec  = 0;
   timer.it_interval.tv_usec = 0;
   timer.it_value.tv_sec     = 0;
   timer.it_value.tv_usec    = 0;
   setitimer(ITIMER_PROF, &timer, NULL);
}


void Quit_Emulator (void)
{
   _go32_dpmi_registers Regs;

   if (no_msg)
      exit(1);

   UninstallMouse();
//   UninstallTimer();
   UninstallKeyboard();

   // Uninstall screen
   Regs.x.ax = 0x1C;
   Regs.h.ah = 0x00;
   Regs.h.al = OldVGAMode;
   Regs.x.ss = Regs.x.sp = Regs.x.flags = 0;
   _go32_dpmi_simulate_int(0x10, &Regs);

   DumpMessages();
   Quit_vMac();
}

void Reset_Emulator (void)
{
   MacReset();
}

// VIA Interface Functions

UBYTE VIA_GORA6 (void) // Main/Alternate Screen Buffer
{
#ifdef _VIA_Interface_Debug
   printf("VIA ORA6 attempts to be an input\n");
#endif
   return 0;
}

void  VIA_PORA6 (UBYTE Data)
{
   vPage2 = Data;
}

UBYTE VIA_GORB6 (void) // Video Beam in Display
{
   return 0; // Assume it is
}

void  VIA_PORB6 (UBYTE Data)
{
#ifdef _VIA_Interface_Debug
  printf("VIA ORB6 attempts to be an output\n");
#endif
}
