// PC - Sound Functions

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "Sound.h"
#include "prefs.h"
#include "ram.h"
#include "via.h"
#include "memory2.h"

#ifdef _Sound_Debug
#include <stdio.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <pc.h>
#include <sys/farptr.h>

// Variables
int Installed;
Sound_Ty Sound;
UWORD kSound_Buffer[370];
UBYTE SoundBuffer[22200];            // sufficient for 60 times [ie 1 second]
UBYTE SoundBuffer2[370];

// Constants
#define FALSE                    0
#define TRUE                     (!(FALSE))

#define MIXER_ADDRESS            0x4
#define MIXER_DATA               0x5
#define DSP_RESET                0x6
#define DSP_READ_DATA            0xA
#define DSP_WRITE_DATA           0xC
#define DSP_WRITE_STATUS         0xC
#define DSP_DATA_AVAIL           0xE

#define SET_TIME_CONSTANT        0x40
#define SPEAKER_ON               0xD1
#define SPEAKER_OFF              0xD3
#define DSP_VER                  0xE1

UWORD Base;
UWORD IRQ;
UWORD DMA_Low;
UWORD DMA_High;
          
UWORD Rate;
int   M;

int   InstallSound(void);
int   DetectSB(void);
int   DetectMixer(void);
int   ResetDSP(void);
UBYTE rwDSP(UBYTE value, UBYTE Read);
UBYTE rwMixer(UBYTE index, UBYTE data, UBYTE Read);
void  SetSpeaker(UBYTE OnOff);
void  SetVolume(UBYTE vol);
void  UninstallSound(void);
void  PlayBack(int t2);
void  Fill_Buffer(UBYTE n, UWORD x);
void  Sound_Play(void);

// Detect the Sound Card
int InstallSound (void)
{
   if (thePrefs.Sound) {
      if (DetectSB()) {
         if (DetectMixer()) {
            printf("vMac: Sound card detected and initialized properly.\n");
            return FALSE;
         } else {
            printf ("vMac: Your sound card does not support Sound Blaster Pro 2.0\n");
            return TRUE;
         }
      } else {
         printf("vMac: Sound card not detected.  Turn sound emulation off in ~vMac.CFG~.\n");
         return TRUE;
      }
   } else {
      return FALSE;
   }
}

// Manually Detect the Sound Card's base addr.
int DetectSB (void)
{
   Base = 0x210;
   while (Base != 0x290) {
      if (ResetDSP())
         return 1;
      Base += 0x10;
   }
   return 0;
}

int DetectMixer (void)
{
   (void)rwMixer(0x22, 0, 1);

   rwMixer(0x22, 0, 0);
   delay(200);

   if ((rwMixer(0x22, 0, 1) & 0xee) != 0)
      return 0;

   rwMixer(0x22, 0xee, 0);
   delay(200);

   if ((rwMixer(0x22, 0, 1) & 0xee) != 0xee)
      return 0;

   return (TRUE);
}

// Resets Sound Blaster DSP chip
int ResetDSP (void)
{
   outportb (Base+DSP_RESET, 1);
   delay(100);
   outportb (Base+DSP_RESET, 0);
   delay(100);
   return (inportb(Base+DSP_READ_DATA) == 0xAA);
}

// Write a byte to DSP chip
UBYTE rwDSP (UBYTE value, UBYTE Read)
{
   if (Read) {
      while (!(inportb(Base+DSP_DATA_AVAIL) & 0x80));
      return (inportb(Base+DSP_READ_DATA));
   } else {
     while (inportb(Base+DSP_WRITE_STATUS) & 0x80);
     outportb(Base+DSP_WRITE_DATA, value);
   }
}

// Read/Write a value to the mixer chip [SB Pro and up..]
UBYTE rwMixer (UBYTE index, UBYTE data, UBYTE Read)
{
   if (Read) {
      outportb(Base+MIXER_ADDRESS, index);
      return (inportb(Base+MIXER_DATA));
   } else {
      outportb(Base+MIXER_ADDRESS, index);
      outportb(Base+MIXER_DATA, data);
   }
}

// turn speakers on/off
void SetSpeaker(UBYTE OnOff)
{
   if (OnOff) {
      rwDSP(SPEAKER_ON, 0);
   } else {
      rwDSP(SPEAKER_OFF, 0);
   }
}

// set sound blaster master volume
void SetVolume (UBYTE vol)
{
   rwMixer(0x22, ((vol+vol*16) & 0xFF), 0);
}                 

// Clean up before exiting
void UninstallSound (void)
{
   ResetDSP();
}

static int p=0, t3;

void PlayBack (int t2)
{
   int    BufferLen, t, qq;
   long   Rate, MaxTicks, Speed;
   ldiv_t clocks;
   int    VOff;
   
   MaxTicks = 1193180L;
   if (M==6) {
      Speed = 44100L;        // for high speed sound cards
   } else {
      Speed = 22050L;
   }

   clocks = ldiv(MaxTicks, Speed);
   BufferLen = t2;
   
   VOff = 0;

   outportb(0x43, 0x34);
   outportb(0x40, 0);
   outportb(0x40, 0);

   Rate = 65536 - clocks.quot;

   for (t=1 ; t<BufferLen ; ++t) {
      rwDSP(0x10, 0);
      rwDSP(SoundBuffer[VOff+t], 0);

      Rate = Rate - clocks.quot;
      if (Rate<0) {
         outportb(0x40, 0);
         outportb(0x40, 0);
         Rate = 65536 - clocks.quot;
      }
      outportb(0x43, 4);
   }
}

void Sound_Play(void)
{
   int   t;
   ULONG kSound_Address;
   char  value;

   if (Sound.Enable==0) {
      if (Sound.Buffer==0)
         kSound_Address = kSnd_Alt_Buffer;
      else 
         kSound_Address = kSnd_Main_Buffer;

      for (t=0 ; t<370 ; t++, t3++) {
         SoundBuffer[t3]=SoundBuffer2[t]=0xff-(RAM_GetW(kSound_Address+t) & 0x00ff);
         p=2;
      }

      if (t3==22200) {
         SetVolume(Sound.Volume*2);
         PlayBack(t3);
         t3=0;
         p=0;
      }
   } else {
      if (p==2) {
         SetVolume(Sound.Volume*2);
         PlayBack(t3);
         t3=0;
         p=0;
      }  
   }
}

// VIA Interface Functions
UBYTE VIA_GORA3 (void) // Main/Alternate Sound Buffer
{
   #ifdef _VIA_Interface_Debug
   printf("VIA ORA3 attempts to be an input\n");
   #endif
   return 0;
}

void VIA_PORA3 (UBYTE Data)
{
   Sound.Buffer = Data; // 1 = Main, 0 = Alternate
}

UBYTE VIA_GORA2 (void) // Sound Volume Bit 2
{
   #ifdef _VIA_Interface_Debug
   printf("VIA ORA2 attempts to be an input\n");
   #endif
   return 0;
}

void VIA_PORA2 (UBYTE Data)
{
   if (Data == 0)
      Sound.Volume &= 0x03;
   else
      Sound.Volume |= 0x04;
}

UBYTE VIA_GORA1 (void) // Sound Volume Bit 1
{
   #ifdef _VIA_Interface_Debug
   printf("VIA ORA1 attempts to be an input\n");
   #endif
   return 0;
}

void VIA_PORA1 (UBYTE Data)
{
   if (Data == 0)
      Sound.Volume &= 0x05;
   else
      Sound.Volume |= 0x02;
}

UBYTE VIA_GORA0 (void) // Sound Volume Bit 0
{
   #ifdef _VIA_Interface_Debug
   printf("VIA ORA0 attempts to be an input\n");
   #endif
   return 0;
}

void VIA_PORA0 (UBYTE Data)
{
   if (Data == 0)
      Sound.Volume &= 0x06;
   else
      Sound.Volume |= 0x01;
}

UBYTE VIA_GORB7 (void) // Sound Enable
{
   #ifdef _VIA_Interface_Debug
   printf("VIA ORB7 attempts to be an input\n");
   #endif
   return 0;
}

void  VIA_PORB7 (UBYTE Data)
{
   Sound.Enable = Data; // 0 = Enabled, 1 = Disabled
}
